//==================================================================
//==================================================================

#include "common.h"
using namespace std;

//==================================================================
// Matrix4 Class Methods
//==================================================================


//------------------------------------------------------------------
// Comparisons
//------------------------------------------------------------------

bool Matrix4::operator==( Matrix4 Q)
{
	
	int i,j;
	for (i=0; i < MATRIX_DIM; i++)
		for (j=0; j < MATRIX_DIM; j++)
			if (
				Q.get(i,j) != 
				v[i][j]
			)
				return false;
	
	return true;
	
}

bool Matrix4::operator!=( Matrix4 other)
{
	return !(*this == other);
}

double& Matrix4::operator[](const int i){
	return *v[i];
}

//returns a flat 16x16, column major array of values for OpenGL
double* Matrix4::GL_matrix(){
	
	Matrix4 transposed = (*this).transpose();
	return transposed.flatten();
	
}

double *Matrix4::flatten(){
	
	//just return a flat 16x16 matrix
	double * nv = (double*)malloc(sizeof(double) * 16);
	int k = 0, i, j;
	
	for (i=0; i < MATRIX_DIM; i++)
		for (j=0; j < MATRIX_DIM; j++){
			nv[k] = v[i][j];
			k++;
		}
			
	return nv;
	
}

//------------------------------------------------------------------
// Compound assignment
//------------------------------------------------------------------

Matrix4& Matrix4::operator*=( Matrix4 m)        // matrix multiplication
{
	double new_v[4][4];
	int i,j, k;
	
	for (i=0; i < MATRIX_DIM; i++)			//row of new
		for (j=0; j < MATRIX_DIM; j++){		//col of new
			//do a dot product of this.row-i * m.col-j
			double t = 0;
			for (k=0; k < MATRIX_DIM; k++){
				t += v[i][k] * m.get(k,j);
			}
			new_v[i][j] = t;
		}

	this->assign(new_v);
	return *this;
}

Matrix4& Matrix4::operator*=( double d)        // scalar mult
{
	double new_v[4][4];
	int i,j, k;
	
	for (i=0; i < MATRIX_DIM; i++)			//row of new
		for (j=0; j < MATRIX_DIM; j++){		//col of new
			new_v[i][j] = v[i][j] * d;
		}

	this->assign(new_v);
	return *this;
}


Matrix4& Matrix4::operator+=( Matrix4 m)        // matrix addition
{
	double new_v[4][4];
	int i,j, k;
	
	for (i=0; i < MATRIX_DIM; i++)			//row of new
		for (j=0; j < MATRIX_DIM; j++){		//col of new
			new_v[i][j] = v[i][j] + m.get(i,j);
		}

	this->assign(new_v);
	return *this;
}

Matrix4& Matrix4::operator-=( Matrix4 m)        // matrix sub
{
	double new_v[4][4];
	int i,j, k;
	
	for (i=0; i < MATRIX_DIM; i++)			//row of new
		for (j=0; j < MATRIX_DIM; j++){		//col of new
			new_v[i][j] = v[i][j] - m.get(i,j);
		}

	this->assign(new_v);
	return *this;
}


//------------------------------------------------------------------
// Matrix4 arithmetic Operations
//------------------------------------------------------------------

Matrix4 Matrix4::operator*(Matrix4 other) {
    return Matrix4(*this) *= other;
}

Matrix4 Matrix4::operator*(double other) {
    return Matrix4(*this) *= other;
}

Matrix4 Matrix4::operator+(Matrix4 other) {
    return Matrix4(*this) += other;
}

Matrix4 Matrix4::operator-(Matrix4 other) {
    return Matrix4(*this) -= other;
}

Vector4& Matrix4::operator*(Vector4 m)      // vector x mat mult
{
	Vector4 *new_v = new Vector4();
	int i,k;
	
	for (i=0; i < MATRIX_DIM; i++)			//rows of resulting vector
		for (k=0; k < MATRIX_DIM; k++){
			(*new_v)[i] += v[i][k] * m[k];
	}

	return *new_v;
}

Vector3& Matrix4::operator*(Vector3 m)      // vector x mat mult
{
	//promote to Vector4
	Vector4 promot_m(m[0], m[1], m[2], 0);
	
	//do the multiplication
	Vector4 new_v = Matrix4(*this) * promot_m;
	//then normalize
	new_v.normalize();
	//demote to V3
	Vector3 *return_v = new Vector3(new_v);
	//and return
	return *return_v;
}


Matrix4& Matrix4::make_ident(){
	
	(*this).set_to(0);
	
	v[0][0] = 1;
	v[1][1] = 1;
	v[2][2] = 1;
	v[3][3] = 1;
	
	return *this;
}


Matrix4& Matrix4::transpose(){
	
	Matrix4 *new_m = new Matrix4();
	int i,j;
	
	for (i=0; i < MATRIX_DIM; i++)			//row of new
		for (j=0; j < MATRIX_DIM; j++){		//col of new
			new_m->set(i, j, v[j][i] );
		}

	return *new_m;
}

//------------------------------------------------------------------
//  Rotations
//------------------------------------------------------------------
Matrix4& Matrix4::make_rotX(double rad){
	
	(*this).set_to(0);
	
	v[0][0] = v[3][3] = 1;
	
	v[1][1] = cos(rad);
	v[1][2] = -sin(rad);
	
	v[2][1] = sin(rad);
	v[2][2] = cos(rad);
	
	return *this;
}

Matrix4& Matrix4::make_rotY(double rad){
	
	(*this).set_to(0);
	
	v[1][1] = v[3][3] = 1;
	
	v[0][0] = cos(rad);
	v[0][2] = sin(rad);
	
	v[2][0] = -sin(rad);
	v[2][2] = cos(rad);
	
	return *this;
}

Matrix4& Matrix4::make_rotZ(double rad){
	
	(*this).set_to(0);
	
	v[2][2] = v[3][3] = 1;
	
	v[0][0] = cos(rad);
	v[0][1] = -sin(rad);
	
	v[1][0] = sin(rad);
	v[1][1] = cos(rad);
	
	return *this;
}

Matrix4& Matrix4::make_rot(double rad_x, double rad_y, double rad_z){
	
	(*this).set_to(0);
	
	Matrix4 x, y, z;
	
	x.make_rotX(rad_x);
	y.make_rotY(rad_y);
	z.make_rotZ(rad_z);
	
	//just multiply them all together
	Matrix4 res = x * y * z;
	
	(*this).assign(res.v);
	
	return *this;
}

Matrix4& Matrix4::make_rot(Vector3 m, double rad){
	
	(*this).set_to(0);
	m.normalize();
	
	
	v[3][3] = 1;
	
	v[0][0] = 1 + (1 - cos(rad)) * (m[0]*m[0] - 1);
	v[0][1] = (-m[2]) * sin(rad) + (1 - cos(rad)) * m[0] * m[1];
	v[0][2] = m[1] * sin(rad) + (1 - cos(rad)) * m[0] * m[2];
	
	v[1][0] = m[2] * sin(rad) + (1 - cos(rad)) * m[1] * m[0];
	v[1][1] = 1 + (1 - cos(rad)) * (m[1]*m[1] - 1);
	v[1][2] = (-m[0]) * sin(rad) + (1 - cos(rad)) * m[1] * m[2];

	v[2][0] = (-m[1]) * sin(rad) + (1 - cos(rad)) * m[2] * m[0];
	v[2][1] = m[0] * sin(rad) + (1 - cos(rad)) * m[2] * m[1];
	v[2][2] = 1 + (1 - cos(rad)) * (m[2]*m[2] - 1);
	
	
	return *this;
}

//scaling, translation

Matrix4& Matrix4::make_scaling(double sx, double sy, double sz){
	(*this).set_to(0);
	
	v[0][0] = sx;
	v[1][1] = sy;
	v[2][2] = sz;
	v[3][3] = 1;
	
	return *this;
	
}


Matrix4& Matrix4::make_translation(double tx, double ty, double tz){
	
	(*this).make_ident();

	v[0][3] = tx;
	v[1][3] = ty;
	v[2][3] = tz;
		
	return *this;
	
}


//------------------------------------------------------------------
//  Special
//------------------------------------------------------------------

// Write output Matrix
std::ostream& operator<<(std::ostream &out, Matrix4 &val) {
	out << "(";

	int i,j;
	for (i=0; i < MATRIX_DIM; i++){
		cout << "{";
		for (j=0; j < MATRIX_DIM; j++){
			cout << val.get(i,j) << ",";
		}
		cout << "}";
	}		

	out << ")";
	return out;
}

