//==================================================================
//==================================================================

#include "common.h"
using namespace std;

#define NORMALS 1
#define COLORFUL 0


//==================================================================
// use a piecewise cubic Bezier curve as the generatrix to calculate 
// a surface revolution 
//==================================================================


Vertex3D *calcSurfaceRevolutionVx(int n, Vector3 CtrlPoints[], int n_points, double angle_delta){
	int i;
	Point2D *CtrlPs = (Point2D*)malloc(sizeof(Point2D) * 3*n+1);
	
	for (i = 0; i< 3*n+1; i++){
		CtrlPs[i].x = CtrlPoints[i][0];
		CtrlPs[i].y = CtrlPoints[i][1];
	}
	
	return calcSurfaceRevolutionVx(n, CtrlPs, n_points, angle_delta);
}

/*
struct Vertex3D    // generic 3D vertex
{
  	Color color;
  	Vector3 coord;
  	Vector3 normal;
	double u,w;		//texture coords
};
struct SurfaceRev
{
	Vertex3D 	vertices[];		//the vertices, includes coords, normal and tex coords
	int 		indexArray[];	//index array of connected vertices
};
struct TrianglePtr
{
	Vertex3D 	*a,*b,*c;		//pointers to the three vertices forming this triangle
};

*/

//n is the number of bezier segments, so we have 3N+1 control points from p0-p3N in the array
//n_points is number of points to evaluate along curve, ie. vertices to calculate
//thus the resulting SurfaceRev will have (360/angle_delta * n_points) vertices in it
//returns an array of vertices, they are stacked from 0,0
Vertex3D *calcSurfaceRevolutionVx(int n, Point2D CtrlPoints[], int n_points, double angle_delta){
	
	int n_verts = (360/angle_delta) * n_points, i, j;
	int nxtVxIdx = 0, nxTriaIdx = 0;
	Vector3 currNormal;
	
	Point2D *linePoints, currLPoint;
	Quad2D *lineQuads;
	Vertex3D *vertices = (Vertex3D*)malloc(sizeof(Vertex3D) * n_verts);
	
	//first we need to calculate the line, use BernsteinCalc
	//or poly if we need to calculate normals
	if (!NORMALS)
		linePoints = piecewiseBernsteinCalc(n, CtrlPoints, n_points);
	else
		lineQuads = piecewisePolyCalc(n, CtrlPoints, n_points);
		
	//the highest y point of the line is its height
	
	//now for each point on the line, we rotate it a full 360
	//which means the vertices in the array will be stacked starting from 0,0
	for (i = 0; i < n_points; i++){
		
		if (!NORMALS){
			currLPoint = linePoints[i];
		}else{
			currNormal[0] = -lineQuads[i].d.y;
			currNormal[1] = lineQuads[i].d.x;
			currNormal[2] = 0;
			currLPoint = lineQuads[i].p;
		}
			
		for (j = 0; j < 360; j += angle_delta){
			//f(u) is x-coordinate
			//x = f(u)sin(phi)
			vertices[nxtVxIdx].coord[0] = currLPoint.x * sin(deg2rad(j));
			vertices[nxtVxIdx].normal[0] = currNormal[0] * sin(deg2rad(j));
			
			//y is the linear param on the y-axis
			vertices[nxtVxIdx].coord[1] = currLPoint.y;
			vertices[nxtVxIdx].normal[1] = currNormal[1];
			
			//z = f(u)cos(phi)
			vertices[nxtVxIdx].coord[2] = currLPoint.x * cos(deg2rad(j));
			vertices[nxtVxIdx].normal[2] = currNormal[0] * cos(deg2rad(j));
		
			//normalize normal
			vertices[nxtVxIdx].normal.normalize();
			
			
			//now set some color
			if (COLORFUL){
				vertices[nxtVxIdx].color.r = sin(((1.0/n_points)*i)*M_PI);
				vertices[nxtVxIdx].color.g = sin(((angle_delta/360)*(j/angle_delta))*M_PI);
				vertices[nxtVxIdx].color.b = (1.0 / n_points / (angle_delta/360)) * (j/angle_delta) * i;
			}else{
				vertices[nxtVxIdx].color.r = 1;
				vertices[nxtVxIdx].color.g = 1;
				vertices[nxtVxIdx].color.b = 1;
			}
			
			//set u,v tex coords
			vertices[nxtVxIdx].u = (double)i / n_points;
			vertices[nxtVxIdx].v = (double)j / 360;
			
			nxtVxIdx++;
		}
	}
	
	return vertices;
}

TrianglePtr *calcSurfaceRevolutionTria(int n, Vector3 CtrlPoints[], int n_points, double angle_delta){
	int i;
	Point2D *CtrlPs = (Point2D*)malloc(sizeof(Point2D) * 3*n+1);
	
	for (i = 0; i< 3*n+1; i++){
		CtrlPs[i].x = CtrlPoints[i][0];
		CtrlPs[i].y = CtrlPoints[i][1];
	}
	
	return calcSurfaceRevolutionTria(n, CtrlPs, n_points, angle_delta);
}

//returns a pointer to an array of triangle pointers, each having 3 ptrs to the forming vertices
TrianglePtr *calcSurfaceRevolutionTria(int n, Point2D CtrlPoints[], int n_points, double angle_delta){

	cout << "calcSurfaceRevolutionTria() called!" << endl;
	int angles = (360/angle_delta);
	int n_triads = (angles * (n_points-1)) * 2; 
	int i, j;
	int nxTriaIdx = 0;
	int y_idx, x_idx;
	
	//alloc some space for the return array
	TrianglePtr *TPtrs = (TrianglePtr*)malloc(sizeof(TrianglePtr) * (n_triads));

	//get the vertices
	Vertex3D *vertices = calcSurfaceRevolutionVx(n, CtrlPoints, n_points, angle_delta);
	Vertex3D *a,*b,*c,*d;
	
	//and now connect them into triangles
	
	for (i = 0; i < n_points-1; i++){
		y_idx = i*angles;
		//cout << "i#" << i << ", yIdx=" << y_idx << "" << endl;
		
		for (j = 0; j < angles; j++){
			x_idx = y_idx + j;
			//cout << "i#" << i << ",j#" << j << ",xIdx=" << x_idx << "" << endl;
			
			//first get the 4 points for the current quad
			//be careful for last set of vertices after full revolution, because the are at index 0, since they wrap around
			//cout << "Assigning a.idx=" << x_idx << ", b.idx=" << (x_idx + angles) << endl;
			
			a = &vertices[x_idx];
			b = &vertices[x_idx + angles];
			
			//as long as we are not on the last face, that is the last quad for this revolution
			if (j < (angles-1)){
				//cout << "Assigning c.idx=" << (x_idx+ angles + 1) << ", d.idx=" << (x_idx + 1) << endl;
				c = &vertices[x_idx + angles + 1];
				d = &vertices[x_idx + 1];
			}else{//last angle, right vertices is of index 0 again
				//cout << "Last angle: Assigning c.idx=" << (y_idx + angles) << ", d.idx=" << (y_idx) << endl;
				c = &vertices[y_idx + angles];
				d = &vertices[y_idx];
			}
			
			//then assign them
			TPtrs[nxTriaIdx].a = a;
			TPtrs[nxTriaIdx].b = b;
			TPtrs[nxTriaIdx].c = c;
			
			nxTriaIdx++;
			TPtrs[nxTriaIdx].a = a;
			TPtrs[nxTriaIdx].b = c;
			TPtrs[nxTriaIdx].c = d;
			nxTriaIdx++;
			
			
		}
		
	}
	
	cout << "calcSurfaceRevolutionTria() returning: " << TPtrs << endl;
	return TPtrs;
	
}

/*	----------------------------------------------------------
*	Calculate line segments using Bernstein Polys
*	---------------------------------------------------------
*/

//returns a piecewise bezier curve, that is an array of vertices for it 
//n is the number of bezier segments, so we have 3N+1 control points from p0-p3N in the array
//n_points is number of points to evaluate along curve, ie. vertices to calculate
//shoud be evenly divisible by n
Point2D *piecewiseBernsteinCalc(int n, Point2D CtrlPoints[], int n_points){

	//the number of points per line segment, that is each piece of the curve
	//subtract 1 to make sure that we run t from 0..1 including 1
	int PsPerPiece = (n_points-1) / n;
	Point2D currCtrlPs[4];
	
	//Point2D vertices[n_points];
	Point2D *vertices = (Point2D*)malloc(sizeof(Point2D) * n_points);
	
	//the running param u: 0<u<3n
	double u, t, nextX = 0;
	int i,j, nextIdx = 0;
	
	//now evaluate each line segment
	for ( i = 0; i < n; i++ ){
		
		currCtrlPs[0] = CtrlPoints[i*3];
		currCtrlPs[1] = CtrlPoints[i*3 + 1];
		currCtrlPs[2] = CtrlPoints[i*3 + 2];
		currCtrlPs[3] = CtrlPoints[i*3 + 3];
		
		//because we run j from 0..psPerPiece, t now runs from 0..1
		//and each point along that segment
		for (j = 0; j <= PsPerPiece; j++){
			//the running t 0..1
			t = (double) j / PsPerPiece;
			
			//the coord
			vertices[nextIdx] = bernsteinCalculat0r(t, currCtrlPs);
			nextIdx++;
		} 
	}
	
	return vertices;
}



//cubic 2D bernstein calculat0r
//gets an array of the 4 control points p0-p4 in the xy-plane and the param t
//and returns x(t)
Point2D bernsteinCalculat0r(double t, Point2D ctrlPs[4]){
	
	int i;
	Point2D ret;
	ret.x = 0; ret.y = 0;
	
	for (i = 0; i < 4; i++){
		double b_t = b_x_t(i, t);
		ret.x += ctrlPs[i].x * b_t;
		ret.y += ctrlPs[i].y * b_t;
	}
	return ret;
	
}

//cubic bernstein polynomial calculation function
double b_x_t(int order, double t){
	double ret;
	
	switch (order){
		case 0:
			ret = -pow(t, 3) + (3 * t*t) - (3*t) + 1;
			break;
		case 1:
			ret = (3 * pow(t, 3)) - (6 * t*t) + (3*t);
			break;
		case 2:
			ret = (-3 * pow(t, 3)) + (3 * t*t);
			break;
		case 3:
			ret = pow(t, 3);
			break;
		default:
			ret = 0;
	}
	
	return ret;
	
}

/*	----------------------------------------------------------
*	Calculate line segments using cubic Polys
*	---------------------------------------------------------
*/

//cubic bernstein polynomial calculation function
//calculates the coeffs a..d
//d=0..a=3
Point2D *PolyCoeffs(Point2D p[4]){
	//Point2D ret[4];
	Point2D *ret = (Point2D*)malloc(sizeof(Point2D) * 4);
	
	ret[3].x = (-p[0].x) + (3*p[1].x) - (3*p[2].x) + p[3].x;
	ret[3].y = (-p[0].y) + (3*p[1].y) - (3*p[2].y) + p[3].y;

	ret[2].x = (3*p[0].x) - (6*p[1].x) + (3*p[2].x);
	ret[2].y = (3*p[0].y) - (6*p[1].y) + (3*p[2].y);

	ret[1].x = (-3*p[0].x) + (3*p[1].x);
	ret[1].y = (-3*p[0].y) + (3*p[1].y);

	ret[0].x = p[0].x;
	ret[0].y = p[0].y;
	
	return ret;
	
}

//cubic 2D Poly calculat0r
//gets an array of the 4 coeffs and param t
//a=3..d=0
//and returns x(t)
Point2D PolyCalculat0r(double t, Point2D coeffs[4]){
	
	int i;
	Point2D ret;
	ret.x = 0; ret.y = 0;
	
	for (i = 0; i < 4; i++){
		ret.x += coeffs[i].x * pow(t,i);
		ret.y += coeffs[i].y * pow(t,i);
	}
	return ret;
	
}

//cubic 2D Poly derivative calculat0r
//gets an array of the 4 coeffs and param t
//a=3..d=0
//and returns x'(t)
Point2D PolyDerivCalculat0r(double t, Point2D coeffs[4]){
	
	int i;
	Point2D ret;
	ret.x = 0; ret.y = 0;
	
	for (i = 1; i < 4; i++){
		ret.x += i * coeffs[i].x * pow(t,i-1);
		ret.y += i * coeffs[i].y * pow(t,i-1);
	}
	return ret;
	
}

//returns a piecewise bezier curve, that is an array of vertices for it 
//this actually returns a quad point, since we also need dx/dy evaluated at each point 
//n is the number of bezier segments, so we have 3N+1 control points from p0-p3N in the array
//n_points is number of points to evaluate along curve, ie. vertices to calculate
//shoud be evenly divisible by n
Quad2D *piecewisePolyCalc(int n, Point2D CtrlPoints[], int n_points){

	//the number of points per line segment, that is each piece of the curve
	int PsPerPiece = n_points / n;
	Point2D currCtrlPs[4];
	Point2D *coeffs;
	
	//Point2D vertices[n_points];
	Quad2D *vertices = (Quad2D*)malloc(sizeof(Quad2D) * n_points);
	
	//the running param u: 0<u<3n
	double u, t, nextX = 0;
	int i,j, nextIdx = 0;
	double a,b,c,d;
	
	//now evaluate each line segment
	for ( i = 0; i < n; i++ ){
		
		currCtrlPs[0] = CtrlPoints[i*3];
		currCtrlPs[1] = CtrlPoints[i*3 + 1];
		currCtrlPs[2] = CtrlPoints[i*3 + 2];
		currCtrlPs[3] = CtrlPoints[i*3 + 3];
		
		//get the coeffs
		coeffs = PolyCoeffs(currCtrlPs);
		
		//and each point along that segment
		for (j = 0; j < PsPerPiece; j++){
			//the running t 0..1
			t = (double) j / PsPerPiece;
			
			//the coord
			vertices[nextIdx].p = PolyCalculat0r(t, coeffs);
			vertices[nextIdx].d = PolyDerivCalculat0r(t, coeffs);
			nextIdx++;
		} 
	}
	
	return vertices;
}


/*	----------------------------------------------------------
*	Calculate line segments using Matrices.. postponed
*	---------------------------------------------------------
*/
